from unittest.mock import patch
from argparse import ArgumentParser
import pytest
from autocommand.autoparse import autoparse, TooManySplitsError


def test_basic_invocation():
    @autoparse
    def func(arg1, arg2: int, opt1=None, opt2=2, opt3='default', flag=False):
        return arg1, arg2, opt1, opt2, opt3, flag

    arg1, arg2, opt1, opt2, opt3, flag = func(
        ['value1', '1', '-o', 'hello', '--opt2', '10', '-f'])

    assert arg1 == 'value1'
    assert arg2 == 1
    assert opt1 == 'hello'
    assert opt2 == 10
    assert opt3 == 'default'
    assert flag is True


def test_invocation_from_argv():
    @autoparse
    def func(arg1, arg2: int):
        return arg1, arg2

    with patch('sys.argv', ['command', '1', '2']):
        arg1, arg2 = func()
        assert arg1 == "1"
        assert arg2 == 2


def test_description_epilog_help(check_help_text):
    @autoparse(
        description='this is a description',
        epilog='this is an epilog')
    def func(arg: 'this is help text'):
        pass

    check_help_text(
        lambda: func(['-h']),
        'this is a description',
        'this is an epilog',
        'this is help text')


def test_docstring_description(check_help_text):
    @autoparse
    def func(arg):
        '''This is a docstring description'''
        pass

    check_help_text(
        lambda: func(['-h']),
        'This is a docstring description')


def test_docstring_description_epilog(check_help_text):
    @autoparse
    def func(arg):
        '''
        This is the description
        -------
        This is the epilog
        '''
        pass

    created_parser = func.parser
    assert 'This is the description' in created_parser.description
    assert 'This is the epilog' in created_parser.epilog

    check_help_text(
        lambda: func(['-h']),
        'This is the description',
        'This is the epilog',
        reject='-------')


def test_bad_docstring():
    with pytest.raises(TooManySplitsError):
        @autoparse
        def func(arg):
            '''
            Part 1
            -------
            Part 2
            -------
            Part 3
            '''
            pass


def test_custom_parser():
    parser = ArgumentParser()

    parser.add_argument('arg', nargs='?')
    group = parser.add_mutually_exclusive_group()
    group.add_argument('-v', '--verbose', action='store_true')
    group.add_argument('-q', '--quiet', action='store_true')

    @autoparse(parser=parser)
    def main(arg, verbose, quiet):
        return arg, verbose, quiet

    assert main([]) == (None, False, False)
    assert main(['thing']) == ('thing', False, False)
    assert main(['-v']) == (None, True, False)
    assert main(['-q']) == (None, False, True)
    assert main(['-v', 'thing']) == ('thing', True, False)

    with pytest.raises(SystemExit):
        main(['-v', '-q'])



from inspect import signature, Signature, Parameter
import pytest
from autocommand.autoparse import KWArgError, PositionalArgError, make_parser


def test_kwargs(check_parse):
    def func(**kwargs): pass

    with pytest.raises(KWArgError):
        make_parser(signature(func), "", "", False)


def test_positional(check_parse):
    # We have to fake this one, because it isn't possible to create a
    # positional-only parameter in pure python

    with pytest.raises(PositionalArgError):
        make_parser(
            Signature([Parameter('arg', Parameter.POSITIONAL_ONLY)]),
            "", "", False)



# Copyright 2014-2016 Nathan West
#
# This file is part of autocommand.
#
# autocommand is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# autocommand is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with autocommand.  If not, see <http://www.gnu.org/licenses/>.

import pytest
from autocommand.automain import automain, AutomainRequiresModuleError


@pytest.mark.parametrize('module_name', ['__main__', True])
def test_name_equals_main_or_true(module_name):
    with pytest.raises(SystemExit):
        @automain(module_name)
        def main():
            return 0


def test_name_not_main_or_true():
    def main():
        return 0

    wrapped_main = automain('some_module')(main)

    assert wrapped_main is main


def test_invalid_usage():
    with pytest.raises(AutomainRequiresModuleError):
        @automain
        def main():
            return 0


def test_args():
    main_called = False
    with pytest.raises(SystemExit):
        @automain(True, args=[1, 2])
        def main(a, b):
            nonlocal main_called
            main_called = True
            assert a == 1
            assert b == 2
    assert main_called


def test_args_and_kwargs():
    main_called = False
    with pytest.raises(SystemExit):
        @automain(True, args=[1], kwargs={'b': 2})
        def main(a, b):
            nonlocal main_called
            main_called = True
            assert a == 1
            assert b == 2

    assert main_called



from autocommand.autoparse import smart_open


def test_smart_open_is_exported():
    import autocommand

    assert autocommand.smart_open is smart_open


def test_smart_open_path_read(tmpdir):
    target = tmpdir.join('file.txt')
    target.write("Hello")

    with smart_open(str(target)) as file:
        assert not file.closed
        assert file.read() == "Hello"

    assert file.closed


def test_smart_open_path_write(tmpdir):
    target = tmpdir.join('file.txt').ensure(file=True)

    with smart_open(str(target), 'w') as file:
        assert not file.closed
        file.write("Test Content")  # Tests that the file is writable

    assert file.closed
    assert target.read() == "Test Content"


def test_smart_open_path_create(tmpdir):
    target = tmpdir.join("file.txt")

    with smart_open(str(target), 'w') as file:
        file.write("Test Content")

    assert target.read() == "Test Content"


def test_smart_open_file(tmpdir):
    path = str(tmpdir.join('file.txt').ensure(file=True))

    with open(path) as file:
        with smart_open(file) as smart_file:
            assert file is smart_file
            assert not smart_file.closed
        assert not smart_file.closed



import pytest


@pytest.mark.parametrize("cli_args", [
    ['-aFoo'],
    ['-a', 'Foo'],
    ['--arg=Foo'],
    ['--arg', 'Foo']
])
def test_optional_flag_styles(check_parse, cli_args):
    check_parse(
        lambda arg="default_value": None,
        *cli_args,
        arg="Foo")


def test_capitalizer(check_parse):
    check_parse(
        lambda arg1="", arg2="", arg3="": None,
        '-a', 'value1', '-A', 'value2', '--arg3', 'value3',
        arg1='value1', arg2='value2', arg3='value3')


def test_reverse_capitalizer(check_parse):
    check_parse(
        lambda Arg1="", Arg2="", Arg3="": None,
        '-A', 'value1', '-a', 'value2', '--Arg3', 'value3',
        Arg1='value1', Arg2='value2', Arg3='value3')


def test_single_letter_param(check_parse):
    def func(a=''):
        pass

    check_parse(
        func,
        '-a', 'value',
        a='value')

    with pytest.raises(SystemExit):
        check_parse(
            func,
            '--a', 'value',
            a='value')


def test_single_letter_prioritized(check_parse):
    '''
    Check that, when deciding which flags get to have single-letter variants,
    single-letter function parameters are assigned a letter first, so they
    aren't recapitalized or forced to have double-dash flags.
    '''
    check_parse(
        lambda arg1='', arg2='', a='': None,
        '-a', 'value1', '-A', 'value2', '--arg2', 'value3',
        a='value1', arg1='value2', arg2='value3')


def test_h_reserved(check_parse):
    check_parse(
        lambda hello='': None,
        '-H', 'value',
        hello='value')



# Copyright 2014-2016 Nathan West
#
# This file is part of autocommand.
#
# autocommand is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# autocommand is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with autocommand.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import sys
from unittest.mock import patch, sentinel
from autocommand import autocommand


autocommand_module = sys.modules['autocommand.autocommand']


def _asyncio_unavailable():
    try:
        import asyncio

        # This is here to silence flake8 complaining about "unused import"
        assert asyncio
    except ImportError:
        return True
    else:
        return False


skip_if_async_unavailable = pytest.mark.skipif(
    _asyncio_unavailable(),
    reason="async tests require asyncio (python3.4+)")


@pytest.fixture
def patched_autoparse():
    with patch.object(
            autocommand_module,
            'autoparse',
            autospec=True) as autoparse:
        yield autoparse


@pytest.fixture
def patched_autoasync():
    with patch.object(
            autocommand_module,
            'autoasync',
            create=True) as autoasync:
        if sys.version_info < (3, 4):
            autoasync.side_effect = NameError('autoasync')

        yield autoasync


@pytest.fixture
def patched_automain():
    with patch.object(
            autocommand_module,
            'automain',
            autospec=True) as automain:

        yield automain


def test_autocommand_no_async(
        patched_automain,
        patched_autoasync,
        patched_autoparse):

    autocommand_wrapped = autocommand(
        sentinel.module,
        description=sentinel.description,
        epilog=sentinel.epilog,
        add_nos=sentinel.add_nos,
        parser=sentinel.parser)(sentinel.original_function)

    assert not patched_autoasync.called

    patched_autoparse.assert_called_once_with(
        sentinel.original_function,
        description=sentinel.description,
        epilog=sentinel.epilog,
        add_nos=sentinel.add_nos,
        parser=sentinel.parser)

    autoparse_wrapped = patched_autoparse.return_value

    patched_automain.assert_called_once_with(sentinel.module)
    patched_automain.return_value.assert_called_once_with(autoparse_wrapped)

    automain_wrapped = patched_automain.return_value.return_value
    assert automain_wrapped is autocommand_wrapped


@pytest.mark.parametrize(
    'input_loop, output_loop',
    [(sentinel.loop, sentinel.loop),
     (True, None)])
@skip_if_async_unavailable
def test_autocommand_with_async(
        patched_automain,
        patched_autoasync,
        patched_autoparse,
        input_loop, output_loop):

    autocommand_wrapped = autocommand(
        sentinel.module,
        description=sentinel.description,
        epilog=sentinel.epilog,
        add_nos=sentinel.add_nos,
        parser=sentinel.parser,
        loop=input_loop,
        forever=sentinel.forever,
        pass_loop=sentinel.pass_loop)(sentinel.original_function)

    patched_autoasync.assert_called_once_with(
        sentinel.original_function,
        loop=output_loop,
        forever=sentinel.forever,
        pass_loop=sentinel.pass_loop)
    autoasync_wrapped = patched_autoasync.return_value

    patched_autoparse.assert_called_once_with(
        autoasync_wrapped,
        description=sentinel.description,
        epilog=sentinel.epilog,
        add_nos=sentinel.add_nos,
        parser=sentinel.parser)
    autoparse_wrapped = patched_autoparse.return_value

    patched_automain.assert_called_once_with(sentinel.module)
    patched_automain.return_value.assert_called_once_with(autoparse_wrapped)
    automain_wrapped = patched_automain.return_value.return_value
    assert automain_wrapped is autocommand_wrapped


def test_autocommand_incorrect_invocation_no_parenthesis(
        patched_automain,
        patched_autoparse,
        patched_autoasync):
    '''
    Test that an exception is raised if the autocommand decorator is called
    without parenthesis by accident
    '''

    with pytest.raises(TypeError):
        @autocommand
        def original_function():
            pass



import pytest
from autocommand.autoparse import AnnotationError


@pytest.mark.parametrize("type_object", [
    int,
    lambda value: "FACTORY({})".format(value)
])
def test_all_annotation_types(check_parse, check_help_text, type_object):
    #  type_object is either `int` or a factory function that converts "str" to
    #  "FACTORY(str)"
    def func(
        typed_arg: type_object,
        note_arg: "note_arg description",
        note_type: ("note_type description", type_object),
        type_note: (type_object, "type_note description")): pass

    check_help_text(
        lambda: check_parse(func, '-h'),
        "note_arg description",
        "note_type description",
        "type_note description")

    check_parse(
        func,
        "1", "2", "3", "4",
        typed_arg=type_object("1"),
        note_arg="2",
        note_type=type_object("3"),
        type_note=type_object("4"))


@pytest.mark.parametrize('bad_annotation', [
    1000,  # An int? What?
    {'foo': 'bar'},  # A dict isn't any better
    [int, 'fooo'],  # For implementation ease we do ask for a tuple
    (),  # Though the tuple should have things in it
    (int,),  # More things
    (int, 'hello', 'world'),  # TOO MANY THINGS
    (int, int),  # The wrong kinds of things
    ("hello", "world"),  # Nope this is bad too
    (lambda value: value, lambda value: value),  # Too many lambdas
])
def test_bad_annotation(bad_annotation, check_parse):
    def func(arg: bad_annotation): pass

    with pytest.raises(AnnotationError):
        check_parse(func)



import pytest


def test_basic_positional(check_parse):
    def func(arg): pass

    check_parse(
        func,
        "foo",
        arg="foo")


@pytest.mark.parametrize('cli_args', [
    [],
    ['value'],
    ['value1', 'value2']
])
def test_variadic_positional(check_parse, cli_args):
    check_parse(
        lambda arg1, *args: None,
        'arg1_value', *cli_args,
        arg1='arg1_value', args=cli_args)


def test_optional_default(check_parse):
    check_parse(
        lambda arg="default_value": None,
        arg="default_value")


@pytest.mark.parametrize('flags, result', [
    ([], False),
    (['-f'], True),
    (['--no-flag'], False),
    (['--no-flag', '-f'], True),
    (['--flag', '--no-flag'], False)
])
def test_add_nos(check_parse, flags, result):
    def func(flag: bool): pass

    check_parse(
        func,
        *flags,
        add_nos=True,
        flag=result)



from inspect import signature
import pytest
from autocommand.autoparse import make_parser


#  TODO: This doesn't really need to be a fixture; replace it with a normal
#  helper
@pytest.fixture(scope='session')
def check_parse():
    def check_parse_impl(function, *args, add_nos=False, **kwargs):
        '''
        Helper for generalized parser testing. This function takes a function,
        a set of positional arguments, and a set of keyword argumnets. It
        creates an argparse parser using `autocommand.autoparse:make_parser` on
        the signature of the provided function. It then parses the positional
        arguments using this parser, and asserts that the returned set of
        parsed arguments matches the given keyword arguments exactly.

        Arguments:
          - function: The function to generate a parser for
          - *args: The set of positional arguments to pass to the generated
            parser
          - add_nos: If True, "-no-" versions of the option flags will be
            created,
            as per the `autoparse` docs.
          - **kwargs: The set of parsed argument values to check for.
        '''
        parser = make_parser(
            func_sig=signature(function),
            description=None,
            epilog=None,
            add_nos=add_nos)

        parsed_args = vars(parser.parse_args(args))
        assert parsed_args == kwargs

    return check_parse_impl


@pytest.fixture
def check_help_text(capsys):
    def check_help_text_impl(func, *texts, reject=()):
        '''
        This helper checks that some set of text is written to stdout or stderr
        after the called function raises a SystemExit. It is used to test that
        the underlying ArgumentParser was correctly configured to output a
        given set of help text(s).

        func: This should be a wrapped autoparse function that causes a
          SystemExit exception to be raised (most commonly a function with the
          -h flag, or with an invalid set of positional arguments). This
          Exception should be accompanied by some kind of printout from
          argparse to stderr or stdout.
        *texts: A set of strings to test for. All of the provided strings will
          be checked for in the captured stdout/stderr using a standard
          substring search.
        reject: A string or set of strings to check do NOT exist anywhere in
          the output. Currently used as a cludge to test the docstring split
          behavior.
        '''
        with pytest.raises(SystemExit):
            func()

        out, err = capsys.readouterr()

        # TODO: be wary of argparse's text formatting
        # TODO: test that the text appears in the order given
        for text in texts:
            assert text in out or text in err

        if isinstance(reject, str):
            reject = [reject]

        for text in reject:
            assert text not in out and text not in err

    return check_help_text_impl



import pytest


def test_int_positional(check_parse):
    def func(arg: int): pass

    check_parse(
        func,
        "1",
        arg=1)

    with pytest.raises(SystemExit):
        check_parse(
            func,
            'hello')


def test_int_default(check_parse):
    def func(arg=10): pass

    check_parse(func, "-a1", arg=1)
    check_parse(func, arg=10)
    with pytest.raises(SystemExit):
        check_parse(func, "-aHello")


def test_int_none_default(check_parse):
    def func(arg: int =None): pass

    check_parse(func, '-a1', arg=1)
    check_parse(func, arg=None)

    with pytest.raises(SystemExit):
        check_parse(func, '-aHello')


# A note on bool types: when the type is EXPLICITLY bool, then the parameter is
# flag no matter what. If the flag is present on the CLI, the default is NOT
# used; this is the most internally consistent behavior (no flag -> default,
# flag -> nondefault). The "truthiness" of the default is used to determine
# the nondefault value- falsy values like `None`, `0`, and `[]` result in True
# being nondefault, while "truthy" values like `1`, `[1]`, and `'hello'` result
# in False being nondefault.
def make_bool_test_params():
    def bool1(flag=False): pass

    def bool2(flag=True): pass

    def bool3(flag: bool): pass

    def bool4(flag: bool =None): pass

    def bool5(flag: bool =0): pass

    def bool6(flag: bool ='noflag'): pass
    return [
        (bool1, True, False),
        (bool2, False, True),
        (bool3, True, False),
        (bool4, True, None),
        (bool5, True, 0),
        (bool6, False, 'noflag')]


@pytest.mark.parametrize(
    'function, with_flag, without_flag',
    make_bool_test_params()
)
def test_bool(check_parse, function, with_flag, without_flag):
    check_parse(function, '-f', flag=with_flag)
    check_parse(function, '--flag', flag=with_flag)
    check_parse(function, flag=without_flag)


def test_file(check_parse, tmpdir):
    filepath = tmpdir.join('test_file.txt')
    filepath.ensure(file=True)
    with filepath.open() as file:
        def func(input_file=file): pass

        check_parse(func, input_file=file)
        check_parse(func, "-i", "path/to/file", input_file="path/to/file")



from setuptools import setup


def getfile(filename):
    with open(filename) as file:
        return file.read()

setup(
    name='autocommand',
    version='2.2.2',
    packages=[
        'autocommand'
    ],
    package_dir={'': 'src'},
    platforms='any',
    license='LGPLv3',
    author='Nathan West',
    url='https://github.com/Lucretiel/autocommand',
    description='A library to create a command-line program from a function',
    long_description=getfile('README.md'),
    classifiers=[
        'Development Status :: 6 - Mature',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



# Copyright 2014-2016 Nathan West
#
# This file is part of autocommand.
#
# autocommand is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# autocommand is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with autocommand.  If not, see <http://www.gnu.org/licenses/>.

import pytest
from contextlib import closing, contextmanager
asyncio = pytest.importorskip('asyncio')
autoasync = pytest.importorskip('autocommand.autoasync').autoasync


class YieldOnce:
    def __await__(self):
        yield


@contextmanager
def temporary_context_loop(loop):
    '''
    Set the given loop as the context loop (that is, the loop returned by
    asyncio.get_event_loop() for the duration of the context)
    '''
    old_loop = asyncio.get_event_loop()
    asyncio.set_event_loop(loop)
    try:
        yield loop
    finally:
        asyncio.set_event_loop(old_loop)


@pytest.fixture
def new_loop():
    '''
    Get a new event loop. The loop is closed afterwards
    '''
    with closing(asyncio.new_event_loop()) as loop:
        yield loop


@pytest.fixture
def context_loop():
    '''
    Create a new event loop and set it as the current context event loop.
    asyncio.get_event_loop() will return this loop within this fixture. Restore
    the original current event loop afterwards. The new loop is also closed
    afterwards.
    '''

    # Can't reuse new_loop() because some tests require new_loop and
    # context_loop to be different
    with closing(asyncio.new_event_loop()) as new_loop:
        with temporary_context_loop(new_loop):
            yield new_loop


def test_basic_autoasync(context_loop):
    data = set()

    async def coro_1():
        data.add(1)
        await YieldOnce()
        data.add(2)

        return 1

    async def coro_2():
        data.add(3)
        await YieldOnce()
        data.add(4)

        return 2

    @autoasync
    async def async_main():
        task1 = asyncio.create_task(coro_1())
        task2 = asyncio.create_task(coro_2())

        result1 = await task1
        result2 = await task2

        assert result1 == 1
        assert result2 == 2

        return 3

    assert async_main() == 3
    assert data == {1, 2, 3, 4}


def test_custom_loop(context_loop, new_loop):
    did_bad_coro_run = False

    async def bad_coro():
        nonlocal did_bad_coro_run
        did_bad_coro_run = True
        await YieldOnce()

    # TODO: this fires a "task wasn't awaited" warning; figure out how to
    # supress
    context_loop.create_task(bad_coro())

    @autoasync(loop=new_loop)
    async def async_main():
        await YieldOnce()
        await YieldOnce()
        return 3

    assert async_main() == 3
    assert did_bad_coro_run is False


def test_pass_loop(context_loop):
    @autoasync(pass_loop=True)
    async def async_main(loop):
        return loop

    assert async_main() is asyncio.get_event_loop()


def test_pass_loop_prior_argument(context_loop):
    '''
    Test that, if loop is the first positional argument, other arguments are
    still passed correctly
    '''
    @autoasync(pass_loop=True)
    async def async_main(loop, argument):
        return loop, argument

    loop, value = async_main(10)
    assert loop is asyncio.get_event_loop()
    assert value == 10


def test_pass_loop_kwarg_only(context_loop):
    @autoasync(pass_loop=True)
    async def async_main(*, loop, argument):
        await YieldOnce()
        return loop, argument

    loop, value = async_main(argument=10)
    assert loop is asyncio.get_event_loop()
    assert value == 10


def test_run_forever(context_loop):
    async def stop_loop_after(t):
        await asyncio.sleep(t)
        context_loop.stop()

    retrieved_value = False

    async def set_value_after(t):
        nonlocal retrieved_value
        await asyncio.sleep(t)
        retrieved_value = True

    @autoasync(forever=True)
    async def async_main():
        asyncio.create_task(set_value_after(0.1))
        asyncio.create_task(stop_loop_after(0.2))
        await YieldOnce()

    async_main()
    assert retrieved_value


def test_run_forever_func(context_loop):
    async def stop_loop_after(t):
        await asyncio.sleep(t)
        context_loop.stop()

    retrieved_value = False

    async def set_value_after(t):
        nonlocal retrieved_value
        await asyncio.sleep(t)
        retrieved_value = True

    @autoasync(forever=True)
    def main_func():
        asyncio.create_task(set_value_after(0.1))
        asyncio.create_task(stop_loop_after(0.2))

    main_func()
    assert retrieved_value


def test_defered_loop(context_loop, new_loop):
    '''
    Test that, if a new event loop is installed with set_event_loop AFTER the
    autoasync decorator is applied (and no loop= is explicitly given to
    autoasync), the new event loop is used when the decorated function is
    called.
    '''
    @autoasync(pass_loop=True)
    async def async_main(loop):
        await YieldOnce()
        return loop

    with temporary_context_loop(new_loop):
        passed_loop = async_main()
        assert passed_loop is new_loop
        assert passed_loop is asyncio.get_event_loop()
        assert passed_loop is not context_loop

    assert passed_loop is not asyncio.get_event_loop()
